from enum import Enum
from typing import Literal, Union

from pydantic import Field
from scale_egp.utils.model_utils import BaseModel, RootModel
from scale_egp.sdk.enums import EmbeddingModelName, EmbeddingConfigType



class EmbeddingConfigBase(BaseModel):
    type: Literal[EmbeddingConfigType.BASE] = Field(
        EmbeddingConfigType.BASE, description="The type of the embedding configuration."
    )
    embedding_model: EmbeddingModelName = Field(
        description="The name of the base embedding model to use. To use custom models, change to type 'models'."
    )


class EmbeddingConfigModelsAPI(BaseModel):
    type: Literal[EmbeddingConfigType.MODELS_API] = Field(
        EmbeddingConfigType.MODELS_API, description="The type of the embedding configuration."
    )
    model_deployment_id: str = Field(
        description="The ID of the deployment of the created model in the Models API V3."
    )


class EmbeddingConfig(RootModel):
    __root__: Union[EmbeddingConfigModelsAPI, EmbeddingConfigBase]


import logging
import sys
from typing import List, Optional
from argh import ArghParser
from scale_egp.cli.collections import (
    CollectionCRUDCommandsForImmutable,
    EGPClientFactory,
    FineTuningJobCommands,
    ModelDeploymentCommands,
    ModelInstanceCommands,
    ModelTemplateCommands,
    TrainingDatasetCommands,
    UserCommands,
    ModelGroupCommands,
)
from scale_egp.cli.formatter import AVAILABLE_FORMATTERS
from scale_egp.cli.parser import get_parser, dispatch


def add_commands_to_parser(parser: ArghParser, commands: CollectionCRUDCommandsForImmutable):
    parser.add_commands(
        [
            getattr(commands, key)
            for key in dir(commands)
            if not key.startswith("_") and callable(getattr(commands, key))
        ],
        group_name=commands.command_group_name,
        group_kwargs={
            "title": getattr(commands, "command_group_title", None),
            "help": getattr(commands, "__doc__", None),
        },
    )


def exec_cli(argv: Optional[List[str]] = None):
    if argv is None:
        argv = sys.argv[1:]
    client_factory = EGPClientFactory()

    parser = get_parser()
    parser.add_argument("--log-curl-commands", action="store_true", default=False)
    parser.add_argument("-k", "--api-key", type=str, default=None, metavar="EGP_API_KEY")
    parser.add_argument(
        "-i",
        "--account-id",
        type=str,
        default=None,
        help="Optional SGP account_id to use. The default account_id will be used if not set.",
        metavar="EGP_ACCOUNT_ID",
    )
    parser.add_argument("-e", "--endpoint-url", type=str, default=None, metavar="EGP_ENDPOINT_URL")
    parser.add_argument(
        "-f", "--format", type=str, choices=[*AVAILABLE_FORMATTERS.keys()], default="rich"
    )
    add_commands_to_parser(parser, ModelGroupCommands(client_factory))
    add_commands_to_parser(parser, ModelInstanceCommands(client_factory))
    add_commands_to_parser(parser, ModelDeploymentCommands(client_factory))
    add_commands_to_parser(parser, ModelTemplateCommands(client_factory))
    add_commands_to_parser(parser, UserCommands(client_factory))
    add_commands_to_parser(parser, FineTuningJobCommands(client_factory))
    add_commands_to_parser(parser, TrainingDatasetCommands(client_factory))
    args = parser.parse_args(argv)
    try:
        client_factory.set_client_kwargs(
            api_key=args.api_key,
            endpoint_url=args.endpoint_url,
            account_id=args.account_id,
            log_curl_commands=args.log_curl_commands,
        )
        if args.log_curl_commands:
            logging.basicConfig(level=logging.INFO)
        return dispatch(parser, args, argv=argv)
    except Exception as e:
        message = getattr(e, "message", None)
        if message is None:
            message = str(e)
        status_code = ""
        if hasattr(e, "code"):
            status_code = f" (HTTP {e.code})"
        print(f"ERROR{status_code}: {message}")
        sys.exit(1)



from __future__ import annotations

from typing import Dict, Optional, Union, Literal, List

from pydantic import Field

from scale_egp.sdk.enums import QuestionType, EvaluationType
from scale_egp.utils.model_utils import BaseModel, RootModel


class CategoricalChoice(BaseModel):
    """
    A choice for a categorical question.

    This is only used in `HUMAN` and `STUDIO` evaluation types to specify a choice for a  question
    that will be asked to users when they are evaluating generated outputs in the SGP annotation UI.

    Attributes:
        label: The text displayed to annotators for this choice.
        value: The value reported in the TestCaseResult for this question if this choice is
            selected.

            If users would like to track the improvement of a model over time, it is
            recommended to use a numeric value for this field.

            A string value can be used if no ordering is desired.
        audit_required: Whether selecting this choice will flag the test case result. Defaulted to false.
    """

    label: str
    value: Union[str, int, bool]
    audit_required: bool = Field(default=False)


# TODO <cathy-scale>: Turn Union[str, List[str]] into a CategoricalRule type when supporting more complex rules
CategoricalCondition = Dict[str, Union[str, List[str]]]
"""
A mapping from `question_id` to either the exact value that must be selected, or a list of
acceptable values. All key-value pairs in the mapping must be satisfied for the condition to be `True`.
For questions with `multi=True`, the selected values include at least one of the acceptable values.
"""


class CategoricalQuestion(BaseModel):
    """
    A categorical question.

    This is only used in `HUMAN` and `STUDIO` evaluation types to specify a choice for a  question
    that will be asked to users when they are evaluating generated outputs in the SGP annotation UI.

    Attributes:
        question_id: A unique id for this question. Each question_id will be represented as a
            key in the `result` field in the TestCaseResult after an annotation is completed.
            When examining how an application has performed on a specific question, users should
            look at the `result` field in the TestCaseResult for the key with the
            corresponding id.
        question_type: The type of the question.
        title: The text displayed to annotators for this question.
        choices: The choices for the question.
        multi: Whether to allow multiple choices to be selected. If `True`, displays as a
            checkbox list. Otherwise, displays as a radio group.
        conditions: Conditions that allow the question to be rendered.
            Example 1: `conditions=[{"accurate": "very", "complete": "yes"}]` means that the selected
            value for `accurate` must be `"very"` and the selected value for `complete` must be
            `"yes"`. Example 2: `conditions=[{"accurate": ["mostly", "very"]}, {"complete": "yes"}]`
            means that either the selected value for `accurate` must be `"mostly"` or `"very"`,
            or the selected value for `complete` must be `"yes"`.
        required: Whether this question is required. If `True`, the annotator must select at least
            one choice.
    """

    question_id: str
    question_type: Literal[QuestionType.CATEGORICAL] = QuestionType.CATEGORICAL.value
    title: str
    choices: List[CategoricalChoice]
    multi: bool = Field(default=False)
    conditions: Optional[List[CategoricalCondition]]
    required: bool = Field(default=False)


class DropdownQuestion(BaseModel):
    """
    A dropdown question.

    This is only used in `HUMAN` and `STUDIO` evaluation types to specify a choice for a  question
    that will be asked to users when they are evaluating generated outputs in the SGP annotation UI.

    Attributes:
        question_id: A unique id for this question. Each question_id will be represented as a
            key in the `result` field in the TestCaseResult after an annotation is completed.
            When examining how an application has performed on a specific question, users should
            look at the `result` field in the TestCaseResult for the key with the
            corresponding id.
        question_type: The type of the question.
        title: The text displayed to annotators for this question.
        choices: The choices for the question.
        conditions: Conditions that allow the question to be rendered.
            Example 1: `conditions=[{"accurate": "very", "complete": "yes"}]` means that the selected
            value for `accurate` must be `"very"` and the selected value for `complete` must be
            `"yes"`. Example 2: `conditions=[{"accurate": ["mostly", "very"]}, {"complete": "yes"}]`
            means that either the selected value for `accurate` must be `"mostly"` or `"very"`,
            or the selected value for `complete` must be `"yes"`.
        required: Whether this question is required. If `True`, the annotator must select at least
            one choice.
    """

    question_id: str
    question_type: Literal[QuestionType.DROPDOWN] = QuestionType.DROPDOWN.value
    title: str
    choices: List[CategoricalChoice]
    conditions: Optional[List[CategoricalCondition]]
    required: bool = Field(default=False)


class FreeTextQuestion(BaseModel):
    """
    A free text question.

    This is only used in `HUMAN` and `STUDIO` evaluation types to specify a choice for a  question
    that will be asked to users when they are evaluating generated outputs in the SGP annotation UI.

    Attributes:
        question_id: A unique id for this question. Each question_id will be represented as a
            key in the `result` field in the TestCaseResult after an annotation is completed.
            When examining how an application has performed on a specific question, users should
            look at the `result` field in the TestCaseResult for the key with the
            corresponding id.
        question_type: The type of the question.
        title: The text displayed to annotators for this question.
        conditions: Conditions that allow the question to be rendered.
            Example 1: `conditions=[{"accurate": "very", "complete": "yes"}]` means that the selected
            value for `accurate` must be `"very"` and the selected value for `complete` must be
            `"yes"`. Example 2: `conditions=[{"accurate": ["mostly", "very"]}, {"complete": "yes"}]`
            means that either the selected value for `accurate` must be `"mostly"` or `"very"`,
            or the selected value for `complete` must be `"yes"`.
        required: Whether this question is required. If `True`, the annotator must input a non-empty string.
    """

    question_id: str
    question_type: Literal[QuestionType.FREE_TEXT] = QuestionType.FREE_TEXT.value
    title: str
    conditions: Optional[List[CategoricalCondition]]
    required: bool = Field(default=False)


class Question(RootModel):
    __root__: Union[CategoricalQuestion, DropdownQuestion, FreeTextQuestion] = Field(
        ...,
        discriminator="question_type",
    )


class StudioEvaluationConfig(BaseModel):
    """
    This specifies the configuration for a studio evaluation job.

    Users should use this evaluation config if they intend to do human evaluation through
    [Studio](https://scale.com/studio).

    Attributes:
        evaluation_type: `EvaluationType.STUDIO`
        studio_project_id: The ID of the Studio project to use for the evaluation.
        questions: The questions to ask users when they are evaluating generated outputs in the
            SGP annotation UI.
    """

    evaluation_type: Literal[EvaluationType.STUDIO] = EvaluationType.STUDIO.value
    studio_project_id: str
    questions: List[Question]


class HumanEvaluationConfig(BaseModel):
    """
    This specifies the configuration for an SGP managed evaluation job.

    Users should use this evaluation config if they intend to do human evaluation through
    the SGP UI.

    Attributes:
        evaluation_type: `EvaluationType.HUMAN`
        questions: The questions to ask users when they are evaluating generated outputs in the
            SGP annotation UI.
    """

    evaluation_type: Literal[EvaluationType.HUMAN] = EvaluationType.HUMAN.value
    questions: List[Question]


class EvaluationConfig(RootModel):
    __root__: Union[StudioEvaluationConfig, HumanEvaluationConfig] = Field(
        ...,
        discriminator="evaluation_type",
    )



from __future__ import annotations

from datetime import datetime
from typing import Optional

from pydantic import Field

from scale_egp.sdk.enums import ModelEndpointType, ModelType
from scale_egp.sdk.types.models import ParameterSchema
from scale_egp.utils.model_utils import Entity, BaseModel

from scale_egp.sdk.models.model_vendor_configuration import ModelVendorConfiguration


class ModelTemplate(Entity):
    """
    This is a template for types of models that can be quickly customized by end users.
    It allows users to upload static docker images that can run specific types of models.
    These docker images will expose parameters that can be injected at ModelAlias creation
    time to customize the functionality. A common example of this is to use a
    HuggingFace LLM template, but swap out model weights for a finetuned model.

    Attributes:
        id: The unique identifier of the entity.
        created_at: The date and time when the entity was created in ISO format.
        account_id: The ID of the account that owns the given entity.
        created_by_user_id: The user who originally created the entity.
        name: The name of the model template
        endpoint_type: The type of endpoint that the model template will create
        model_type: The type of model that the model template will create
        vendor_configuration: The vendor configuration of the model template
        model_creation_parameters_schema: The field names and types of available parameter fields
            which may be specified during model creation
        model_request_parameters_schema: The field names and types of available parameter fields
            which may be specified in a model execution API's `model_request_parameters` field.
    """

    name: str
    endpoint_type: ModelEndpointType
    model_type: ModelType
    vendor_configuration: ModelVendorConfiguration
    model_creation_parameters_schema: Optional[ParameterSchema] = Field(
        None,
        description="The field names and types of available parameter fields which may be "
        "specified during model creation",
    )
    model_request_parameters_schema: Optional[ParameterSchema] = Field(
        None,
        description="The field names and types of available parameter fields which may be "
        "specified in a model execution API's `model_request_parameters` field.",
    )
    id: str = Field(..., description="The unique identifier of the entity.")
    created_at: datetime = Field(
        ..., description="The date and time when the entity was created in ISO format."
    )
    account_id: str = Field(..., description="The ID of the account that owns the given entity.")
    created_by_user_id: str = Field(..., description="The user who originally created the entity.")


class ModelTemplateRequest(BaseModel):
    name: str
    endpoint_type: ModelEndpointType
    model_type: ModelType
    vendor_configuration: ModelVendorConfiguration
    model_creation_parameters_schema: Optional[ParameterSchema] = Field(
        None,
        description="The field names and types of available parameter fields which may be "
        "specified during model creation",
    )
    model_request_parameters_schema: Optional[ParameterSchema] = Field(
        None,
        description="The field names and types of available parameter fields which may be "
        "specified in a model execution API's `model_request_parameters` field.",
    )
    account_id: str = Field(..., description="The ID of the account that owns the given entity.")



from typing import IO, List, Optional
import httpx
from httpx._types import FileTypes

from scale_egp.utils.api_utils import APIEngine
from scale_egp.sdk.models.fine_tuning_jobs import FineTuningJobVendorConfiguration
from scale_egp.sdk.types.fine_tuning import (
    FineTuningJob,
    FineTuningJobRequest,
    TrainingDataset,
    TrainingDatasetORMSchemaTypeEnum,
)
import os.path


class TrainingDatasetCollection(APIEngine):
    """
    Collections class for SGP Model fine tuning jobs.
    """

    _sub_path = "v3/training-datasets"

    def create(
        self,
        file_content: Optional[FileTypes] = None,
        file_name: Optional[str] = None,
        account_id: Optional[str] = None,
        name: Optional[str] = None,
        schema_type: TrainingDatasetORMSchemaTypeEnum = TrainingDatasetORMSchemaTypeEnum.GENERATION,
    ) -> FineTuningJob:
        """
        Create a new SGP Model.

        Returns:
            The created Model.
        """
        if bool(file_content is None) == bool(file_name is None):
            raise Exception("Please specify either file_content or file_name")
        file_handle: Optional[IO[bytes]] = None
        file = file_content
        if file_name:
            file_handle = open(file_name, "rb")
            file = (os.path.basename(file_name), file_handle)
        name = os.path.basename(file_name) if file_name is not None else (name or "unnamed dataset")
        response = self._post(
            sub_path=self._sub_path,
            request={
                "account_id": account_id or self._api_client.account_id,
                "name": name,
                "schema_type": schema_type.value,
            },
            file=file,
        )
        if file_handle:
            file_handle.close()
        return TrainingDataset.from_dict(response.json())

    def get(
        self,
        id: str,
    ) -> FineTuningJob:
        """
        Get a Model fine tuning job by ID.

        Returns:
            The finet tuning job.
        """
        response = self._get(
            sub_path=f"{self._sub_path}/{id}",
        )
        return FineTuningJob.from_dict(response.json())

    def delete(
        self,
        id: str,
    ) -> bool:
        """
        Cancels a fine-tuning job by ID.

        Returns:
            True if the model was successfully deleted.
        """
        response = self._delete(
            sub_path=f"{self._sub_path}/{id}",
        )
        return response.status_code == httpx.codes.ok

    def list(
        self,
    ) -> List[FineTuningJob]:
        """
        List all models that the user has access to.

        Returns:
            A list of Studio Projects.
        """
        response = self._get(
            sub_path=self._sub_path,
        )
        return [FineTuningJob.from_dict(fine_tuning_job) for fine_tuning_job in response.json()]


class FineTuningJobCollection(APIEngine):
    """
    Collections class for SGP Model fine tuning jobs.
    """

    _sub_path = "v3/fine-tuning-jobs"

    def create(
        self,
        base_model_id: str,
        vendor_configuration: Optional[FineTuningJobVendorConfiguration],
        training_dataset_id: str,
        validation_dataset_id: Optional[str],
        account_id: Optional[str],
    ) -> FineTuningJob:
        """
        Create a new SGP Model.

        Returns:
            The created Model.
        """
        response = self._post(
            sub_path=self._sub_path,
            request=FineTuningJobRequest(
                base_model_id=base_model_id,
                vendor_configuration=vendor_configuration,
                training_dataset_id=training_dataset_id,
                validation_dataset_id=validation_dataset_id,
                account_id=account_id or self._api_client.account_id,
            ),
        )
        return FineTuningJob.from_dict(response.json())

    def get(
        self,
        id: str,
    ) -> FineTuningJob:
        """
        Get a Model fine tuning job by ID.

        Returns:
            The finet tuning job.
        """
        response = self._get(
            sub_path=f"{self._sub_path}/{id}",
        )
        return FineTuningJob.from_dict(response.json())

    def delete(
        self,
        id: str,
    ) -> bool:
        """
        Cancels a fine-tuning job by ID.

        Returns:
            True if the model was successfully deleted.
        """
        response = self._delete(
            sub_path=f"{self._sub_path}/{id}",
        )
        return response.status_code == httpx.codes.ok

    def list(
        self,
    ) -> List[FineTuningJob]:
        """
        List all models that the user has access to.

        Returns:
            A list of Studio Projects.
        """
        response = self._get(
            sub_path=self._sub_path,
        )
        return [FineTuningJob.from_dict(fine_tuning_job) for fine_tuning_job in response.json()]



import random
import string

from datetime import datetime
from typing import Optional, Any, Dict, List, Tuple, Union, Literal, TypeVar, Type

from pydantic import Field, Extra, BaseConfig, create_model

from scale_egp.sdk.enums import ModelVendor, AgentAction, ModelType
from scale_egp.sdk.types.agents import ActionContext, Tool
from scale_egp.sdk.types.memory_strategy import MemoryStrategy
from scale_egp.sdk.types.message import Message
from scale_egp.utils.model_utils import Entity, BaseModel
from scale_egp.sdk.models.model_vendor_configuration import DeploymentVendorConfiguration


ParameterValueType = Union[str, int, float, bool]


class ParameterBindings(BaseModel):
    bindings: Dict[str, ParameterValueType]

    class Config:
        extra = Extra.forbid


class ModelInstance(Entity):
    """
    Entity for all models, including both self-hosted and 3rd party, base, and fine-tuned models.

    Attributes:
        id: The unique identifier of the model.
        created_at: The date and time when the entity was created in ISO format.
        account_id: The ID of the account that owns the given entity.
        created_by_user_id: The user who originally created the entity.
        name: The name of the model
        description: The description of the model
        model_vendor: The vendor of the model
        base_model_id: The ID of the base model
        base_model_metadata: Metadata for the base model
        model_template_id: The ID of the model template
        model_group_id: Model group that the entity belongs to
    """

    name: str
    model_type: ModelType
    model_vendor: Optional[ModelVendor] = Field(None)
    model_creation_parameters: Optional[Dict[str, Any]] = Field(None)
    base_model_id: Optional[str] = Field(None)
    base_model_metadata: Optional[Dict[str, Any]] = Field(None)
    description: Optional[str] = Field(None)
    model_template_id: Optional[str] = Field(None)
    id: str = Field(..., description="The unique identifier of the entity.")
    created_at: datetime = Field(
        ..., description="The date and time when the entity was created in ISO format."
    )
    account_id: str = Field(..., description="The ID of the account that owns the given entity.")
    created_by_user_id: str = Field(..., description="The user who originally created the entity.")
    model_group_id: Optional[str] = Field(None)
    training_data_card: Optional[str] = Field(None)
    model_card: Optional[str] = Field(None)
    request_schema: Dict[str, Any]
    response_schema: Dict[str, Any]


class ModelInstanceRequest(BaseModel):
    name: str
    model_type: ModelType
    model_vendor: Optional[ModelVendor] = Field(None)
    model_creation_parameters: Optional[Dict[str, Any]] = Field(None)
    base_model_id: Optional[str] = Field(None)
    base_model_metadata: Optional[Dict[str, Any]] = Field(None)
    description: Optional[str] = Field(None)
    model_template_id: Optional[str] = Field(None)
    account_id: str = Field(..., description="The ID of the account that owns the given entity.")
    model_group_id: Optional[str] = Field(None)
    training_data_card: Optional[str] = Field(None)
    model_card: Optional[str] = Field(None)


class ModelDeployment(Entity):
    """
    Deployment of a model

    Attributes:
        id: The unique identifier of the model deployment.
    """

    id: str
    name: str
    model_creation_parameters: Optional[Dict[str, Any]]
    vendor_configuration: Optional[DeploymentVendorConfiguration] = Field(None)
    model_instance_id: Optional[str]
    status: Optional[str]  # TODO <cathy-scale>: enum
    # TODO <cathy-scale>: put these repeated fields in a parent class
    created_at: datetime = Field(
        ..., description="The date and time when the entity was created in ISO format."
    )
    account_id: str = Field(..., description="The ID of the account that owns the given entity.")
    created_by_user_id: str = Field(..., description="The user who originally created the entity.")


class ModelDeploymentRequest(BaseModel):
    name: str
    vendor_configuration: Optional[DeploymentVendorConfiguration] = Field(None)
    model_creation_parameters: Optional[Dict[str, Any]] = Field(None)
    account_id: str = Field(..., description="The ID of the account that owns the given entity.")


class BaseModelResponse(BaseModel):
    class Config(BaseModel.Config):
        extra = Extra.forbid


class AgentResponse(BaseModelResponse):
    """
    Response schema for agents.

    See the [Execute Agent REST API](https://scale-egp.readme.io/reference/execute_agent) for more
    information.

    Attributes:
        action: The action that the agent performed.
        context: Context object containing the output payload. This will contain a key for all
            actions that the agent can perform. However, only the key corresponding to
            the action that the agent performed have a populated value. The rest of the
            values will be `null`.
    """

    action: AgentAction = Field(...)
    context: ActionContext = Field(...)


class BaseModelRequest(BaseModel):
    model_request_parameters: Optional["ParameterBindings"] = Field(None)

    class Config(BaseModel.Config):
        extra = Extra.forbid


class AgentRequest(BaseModelRequest):
    """
    Response schema for agents.
    See the [Execute Agent REST API](https://scale-egp.readme.io/reference/execute_agent) for more
    information.

    Attributes:
        memory_strategy: The memory strategy to use for the agent. A memory strategy
            is a way to prevent the underlying LLM's context limit from being exceeded.
            Each memory strategy uses a different technique to condense the input message
            list into a smaller payload for the underlying LLM.
        tools: The list of specs of tools that the agent can use. Each spec must contain
            a `name` key set to the name of the tool, a `description` key set to the
            description of the tool, and an `arguments` key set to a JSON Schema
            compliant object describing the tool arguments.

            The name and description of each tool is used by the agent to decide when to
            use certain tools. Because some queries are complex and may require multiple
            tools to complete, it is important to make these descriptions as
            informative as possible. If a tool is not being chosen when it should,
            it is common practice to tune the description of the tool to make it more
            apparent to the agent when the tool can be used effectively.
        messages: The list of messages in the conversation.
        instructions: The initial instructions to provide to the agent.

            Use this to guide the agent to act in more specific ways. For example, if you
            have specific rules you want to restrict the agent to follow you can specify them here.
            For example, if I want the agent to always use certain tools before others,
            I can write that rule in these instructions.

            Good prompt engineering is crucial to getting performant results from the
            agent. If you are having trouble getting the agent to perform well,
            try writing more specific instructions here before trying more expensive
            techniques such as swapping in other models or finetuning the underlying LLM.

    """

    memory_strategy: Optional[MemoryStrategy] = Field(default=None)
    tools: List[Tool] = Field(...)
    messages: List[Message] = Field(...)
    instructions: Optional[str] = Field(
        default="You are an AI assistant that helps users with their questions. You "
        "can answer questions directly or acquire information from any of the "
        "attached tools to assist you. Always answer the user's most recent query to the "
        "best of your knowledge.\n\n"
        "When asked about what tools are available, you must list each attached "
        "tool's name and description. When asked about what you can do, mention "
        "that in addition to your normal capabilities, you can also use the attached "
        "tools by listing their names and descriptions. You cannot use any other tools "
        "other than the ones provided to you explicitly.",
    )


class ChatCompletionRequest(BaseModelRequest):
    """
    Request schema for chat completion models.

    Attributes:
        temperature: What sampling temperature to use, between [0, 1]. Higher values like 0.8
            will make the output more random, while lower values like 0.2 will make it
            more focused and deterministic. Setting temperature=0.0 will enable fully
            deterministic
            (greedy) sampling.
        stop_sequences: List of up to 4 sequences where the API will stop generating further tokens.
            The returned text will not contain the stop sequence.
        max_tokens: The maximum number of tokens to generate in the completion. The token count
            of your prompt plus max_tokens cannot exceed the model's context length. If not,
            specified, max_tokens will be determined based on the model used.
        messages: List of messages for the chat completion to consider when generating a response.
    """

    temperature: float = Field(
        default=0.2,
        ge=0.0,
        le=1.0,
    )
    stop_sequences: Optional[List[str]] = Field(
        default=None,
        max_items=4,
    )
    max_tokens: Optional[int] = Field(
        default=None,
    )
    messages: List[Message]


class ChatCompletionResponse(BaseModelResponse):
    """
    Response schema for chat completion models.

    Attributes:
        message: The generated message from the chat completion model.
        finish_reason: The reason the chat completion finished.
    """

    message: Message
    finish_reason: Optional[str]


class CompletionRequest(BaseModelRequest):
    """
    Request schema for completion models.

    Attributes:
        temperature: What sampling temperature to use, between [0, 1]. Higher values like 0.8
            will make the output more random, while lower values like 0.2 will make it
            more focused and deterministic. Setting temperature=0.0 will enable fully
            deterministic
            (greedy) sampling.
        stop_sequences: List of up to 4 sequences where the API will stop generating further tokens.
            The returned text will not contain the stop sequence.
        max_tokens: The maximum number of tokens to generate in the completion. The token count
            of your prompt plus max_tokens cannot exceed the model's context length. If not,
            specified, max_tokens will be determined based on the model used.
        prompts: List of prompts to generate completions for.
    """
    
    temperature: float = Field(
        default=0.2,
        ge=0.0,
        le=1.0,
    )
    stop_sequences: Optional[List[str]] = Field(
        default=None,
        max_items=4,
    )
    max_tokens: Optional[int] = Field(
        default=None,
    )
    prompts: List[str]


class CompletionResponse(BaseModelResponse):
    """
    Response schema for completion models.

    Attributes:
        completions: List of prompt, completion pairs.
        finish_reason: The reason the completion finished.
    """

    # List of prompt, completion pairs, eg: [("prompt", ["completion1", "completion2", ...]), ...]
    finish_reason: Optional[str]
    prompt_tokens: Optional[int]
    response_tokens: Optional[int]
    completions: List[Tuple[str, List[str]]]


class EmbeddingRequest(BaseModelRequest):
    """
    Request schema for embedding models.

    Attributes:
        texts: List of texts to get embeddings for.
    """

    texts: List[str]


class EmbeddingResponse(BaseModelResponse):
    """
    Response schema for embedding models.

    Attributes:
        embeddings: List of text, embedding pairs.
    """

    # List of text, embedding pairs, eg: [("text to embed", [1.0, 1.05, 2.07, ...]), ...]
    embeddings: List[Tuple[str, List[float]]]


class RerankingRequest(BaseModelRequest):
    """
    Request schema for reranking models.

    Attributes:
        query: Query to rerank chunks against in order of relevance.
        chunks: List of chunks to rerank.
    """

    query: str
    chunks: List[str]


class RerankingResponse(BaseModelResponse):
    """
    Response schema for reranking models.

    Attributes:
        chunk_scores: List of scores for each chunk in the same order as the input chunks.
    """

    chunk_scores: List[float]


def get_random_string(length=8):
    # choose from all lowercase letter
    letters = string.ascii_lowercase
    return "".join(random.choice(letters) for i in range(length))


class ParameterSchemaField(BaseModel):
    """
    The schema used to specify the type and info about each parameter to be passed for model
    creation.

    Attributes:
        name: Name of the parameter.
        type: Type of the parameter.
        description: Description of the parameter.
        required: Whether the parameter is required or not.
    """

    name: str
    type: Union[Literal["str"], Literal["int"], Literal["float"], Literal["bool"]]
    description: str
    required: bool  # default is optional


class ParameterSchema(BaseModel):
    """
    The schema used to specify the parameters to be passed for model creation.

    Attributes:
        parameters: List of parameter schema fields.
    """

    parameters: List[ParameterSchemaField]


BaseModelT = TypeVar("BaseModelT", bound=BaseModel)


class ParameterSchemaModelConfig(BaseConfig):
    extra = Extra.forbid


def parameter_schema_to_model(
    model_name: str, parameter_schema: ParameterSchema
) -> Type[BaseModelT]:
    return create_model(
        model_name,
        __config__=ParameterSchemaModelConfig,
        **{
            field.name: (field.type, ...) if field.required else (field.type, None)
            for field in parameter_schema.parameters
        },
    )



from typing import Literal

from pydantic import BaseModel, Field


# Memory Strategies
class LastKMemoryStrategyParams(BaseModel):
    """
    Parameters for the last k memory strategy

    Attributes:
        k: The maximum number of previous messages to remember.
    """
    k: int = Field(
        ...,
        ge=1,
    )


class LastKMemoryStrategy(BaseModel):
    """
    Last K Memory Strategy. This strategy truncates the message history to the last `k`
    messages. It is the simplest way to prevent the model's context limit from being
    exceeded. However, this strategy only allows the model to have short term memory.
    For longer term memory, please use one of the other strategies.

    Attributes:
        name: Name of the memory strategy. Must be `last_k`.
        params: Configuration parameters for the memory strategy.
    """
    name: Literal["last_k"] = Field(
        default="last_k",
        const=True,
    )
    params: LastKMemoryStrategyParams = Field(...)

    class Config(BaseModel.Config):
        title = "Last K Memory Strategy"


MemoryStrategy = LastKMemoryStrategy



import json
from typing import Any, Dict, Generator, Generic, Optional, Type, TypeVar
from scale_egp.cli.formatter import (
    FormattingOptions,
    get_formatting_options,
    set_formatting_options,
)
from scale_egp.cli.model_instance_description import ModelInstanceDescription
from scale_egp.sdk.client import EGPClient
from scale_egp.sdk.collections.model_templates import ModelTemplateCollection
from scale_egp.sdk.collections.models import ModelDeploymentCollection, ModelInstanceCollection
from scale_egp.sdk.collections.fine_tuning import FineTuningJobCollection, TrainingDatasetCollection
from scale_egp.sdk.enums import ModelVendor
from scale_egp.sdk.types.models_group import ModelGroup, ModelGroupRequest
from scale_egp.sdk.collections.model_groups import ModelGroupCollection
from scale_egp.sdk.types.user_info import UserInfoResponse
from scale_egp.sdk.types.models import (
    ModelDeployment,
    ModelDeploymentRequest,
    ModelInstance,
    ModelInstanceRequest,
)
from scale_egp.sdk.types.fine_tuning import (
    FineTuningJob,
    FineTuningJobRequest,
    TrainingDataset,
    TrainingDatasetRequest,
    TrainingDatasetORMSchemaTypeEnum,
)
from scale_egp.sdk.types.model_templates import ModelTemplate, ModelTemplateRequest
from scale_egp.sdk.constants.model_schemas import MODEL_SCHEMAS
from scale_egp.utils.api_utils import APIEngine
from scale_egp.utils.model_utils import BaseModel
from argh import CommandError, arg
import fastjsonschema
from scale_egp.sdk.models.fine_tuning_jobs import FineTuningJobEvent


EntityT = TypeVar("EntityT", bound=BaseModel)
RequestT = TypeVar("RequestT", bound=BaseModel)


def read_json_file(filename: str, expected_type: Optional[Any] = dict) -> Any:
    with open(filename, "r", encoding="utf-8") as f:
        data = json.loads(f.read())
        if expected_type is not None:
            assert isinstance(data, expected_type)
        return data


class EGPClientFactory:
    def __init__(
        self,
    ):
        self.client: Optional[EGPClient] = None
        self._client_kwargs = None

    def set_client_kwargs(self, **kwargs):
        self._client_kwargs = kwargs

    def get_client(self) -> EGPClient:
        if self.client is None:
            self.client = EGPClient(**self._client_kwargs)
        return self.client


class CollectionCRUDCommandsForImmutable(Generic[EntityT, RequestT]):
    command_group_name = "CRUD"
    command_group_title: Optional[str] = None

    def __init__(
        self,
        client_factory: EGPClientFactory,
        entity_type: Type[EntityT],
        request_type: Type[RequestT],
        collection_type: Type[APIEngine],
    ):
        self._client_factory = client_factory
        self._entity_type = entity_type
        self._request_type = request_type
        self._collection_type = collection_type

    def _get_collection_instance(self) -> APIEngine:
        return self._collection_type(self._client_factory.get_client())

    def _transform_entity_json(self, entity_dict: Dict[str, Any]) -> Dict[str, Any]:
        return entity_dict

    def _get_api_subpath_prefix(self) -> str:
        collection = self._get_collection_instance()
        return getattr(collection, "_sub_path")

    def _create(self, request_dict: Any) -> EntityT:
        # add client account id if not set in file
        request_dict["account_id"] = request_dict.get(
            "account_id", self._client_factory.get_client().account_id
        )
        request_obj = self._request_type(**request_dict)
        collection = self._get_collection_instance()
        response = collection._post(self._get_api_subpath_prefix(), request_obj)
        assert response.status_code == 200
        response_dict = response.json()
        assert isinstance(response_dict, dict)
        return self._entity_type(**response_dict)

    @arg("filename", help="file to load")
    def create(self, filename: str) -> EntityT:
        request_dict = read_json_file(filename)
        return self._create(request_dict)

    def get(self, id: str) -> EntityT:
        collection = self._get_collection_instance()
        sub_path = f"{self._get_api_subpath_prefix()}/{id}"
        response = collection._get(sub_path)
        assert response.status_code == 200
        response_dict = response.json()
        assert isinstance(response_dict, dict)
        return self._entity_type(**response_dict)

    def delete(self, id: str) -> None:
        collection = self._get_collection_instance()
        sub_path = f"{self._get_api_subpath_prefix()}/{id}"
        response = collection._delete(sub_path)
        assert response.status_code == 200

    def list(self) -> Generator[EntityT, None, None]:
        formatting_options = get_formatting_options()
        formatting_options.force_list = True
        set_formatting_options(formatting_options)
        collection = self._get_collection_instance()
        response = collection._get(self._get_api_subpath_prefix())
        assert response.status_code == 200
        response_list = response.json()
        assert isinstance(response_list, list)
        # TODO: pagination
        for entity_dict in response_list:
            yield self._entity_type(**entity_dict)


class CollectionCRUDCommandsWithUpdate(
    Generic[EntityT, RequestT], CollectionCRUDCommandsForImmutable[EntityT, RequestT]
):
    def _update(self, id: str, request_dict: Any) -> EntityT:
        # add client account id if not set in file
        request_dict["account_id"] = request_dict.get(
            "account_id", self._client_factory.get_client().account_id
        )
        request_obj = self._request_type(**request_dict)
        collection = self._get_collection_instance()
        response = collection._patch(f"{self._get_api_subpath_prefix()}/{id}", request_obj)
        assert response.status_code == 200
        response_dict = response.json()
        assert isinstance(response_dict, dict)
        return self._entity_type(**response_dict)

    @arg("filename", help="file to load")
    def update(self, id: str, filename: str) -> EntityT:
        request_dict = read_json_file(filename)
        return self._update(id, request_dict)


class ModelInstanceCommands(CollectionCRUDCommandsWithUpdate[ModelInstance, ModelInstanceRequest]):
    command_group_name = "model-instance"

    def __init__(
        self,
        client_factory: EGPClientFactory,
    ):
        super().__init__(
            client_factory, ModelInstance, ModelInstanceRequest, ModelInstanceCollection
        )

    def list(self) -> Generator[EntityT, None, None]:
        set_formatting_options(
            FormattingOptions(table_columns=["id", "name", "model_template_id", "description"])
        )
        return super().list()

    def describe(self, model_id: str) -> ModelInstanceDescription:
        model_instance = self.get(model_id)
        model_template = ModelTemplateCommands(self._client_factory).get(
            model_instance.model_template_id
        )
        return ModelInstanceDescription(
            model_instance=model_instance, model_template=model_template
        )

    @arg("filename", help="file to load")
    @arg(
        "--model-template-id", help="id of the model template to use if not specified in JSON file"
    )
    def create(self, filename: str, model_template_id: Optional[str] = None) -> EntityT:
        request_dict = read_json_file(filename)
        if request_dict.get("model_vendor", ModelVendor.LAUNCH.value) == ModelVendor.LAUNCH.value:
            effective_model_template_id = request_dict.get("model_template_id", model_template_id)
            if effective_model_template_id is None:
                raise CommandError(
                    "No model template id specified in JSON file or --model-template-id option. Please provide the model template id for Launch models!"
                )
            request_dict["model_template_id"] = effective_model_template_id
            if request_dict.get("model_type") is None:
                # if no model_type is specified in model alias json, use model_type of template
                model_template = ModelTemplateCommands(self._client_factory).get(
                    effective_model_template_id
                )
                request_dict["model_type"] = model_template.model_type

        return self._create(request_dict)

    def _validate_request(self, model_instance: ModelInstance, request: Dict[str, Any]):
        validator = fastjsonschema.compile(model_instance.request_schema)
        return validator(request)

    @arg("filename", help="Model request")
    def validate_request(self, model_instance_id: str, filename: str) -> Optional[str]:
        model_instance = self.get(model_instance_id)
        execute_request_dict = read_json_file(filename)
        try:
            self._validate_request(model_instance, execute_request_dict)
        except fastjsonschema.JsonSchemaException as e:
            return f"Data failed validation: {e}"
        return None


class ModelDeploymentCommands(
    CollectionCRUDCommandsWithUpdate[ModelDeployment, ModelDeploymentRequest]
):
    command_group_name = "model-deployment"

    def __init__(
        self,
        client_factory: EGPClientFactory,
    ):
        super().__init__(
            client_factory, ModelDeployment, ModelDeploymentRequest, ModelDeploymentCollection
        )
        self.model_instance_id = None

    def _get_api_subpath_prefix(self) -> str:
        collection = self._get_collection_instance()
        return getattr(collection, "_sub_path").format(model_id=self.model_instance_id)

    def list(self, model_instance_id: str) -> Generator[EntityT, None, None]:
        self.model_instance_id = model_instance_id
        set_formatting_options(
            FormattingOptions(table_columns=["id", "name", "created_at", "description"])
        )
        return super().list()

    @arg("filename", nargs="?", help="JSON file specifying the deployment configuration")
    @arg("--model-deployment-name", help="Model deployment name")
    @arg("--model-instance-id", help="Model instance for which to create a deployment")
    def create(
        self,
        filename: Optional[str],
        model_instance_id: Optional[str] = None,
        model_deployment_name: Optional[str] = None,
    ) -> EntityT:
        request_dict = read_json_file(filename) if filename else {}
        if model_instance_id is not None:
            request_dict["model_instance_id"] = model_instance_id
        if request_dict.get("model_instance_id") is None:
            raise CommandError(
                "Please specify a model instance id, either using the --model-instance-id command line option or in the provided JSON file"
            )
        self.model_instance_id = request_dict["model_instance_id"]
        if model_deployment_name:
            request_dict["name"] = model_deployment_name
        if request_dict.get("name") is None:
            model_instance = ModelInstanceCommands(self._client_factory).get(
                request_dict["model_instance_id"]
            )
            # default to model instance name for the deployment name if its not set to something else
            request_dict["name"] = model_instance.name
        return self._create(request_dict)

    def get(self, model_instance_id: str, model_deployment_id: str) -> EntityT:
        self.model_instance_id = model_instance_id
        return super().get(model_deployment_id)

    def delete(self, model_instance_id: str, model_deployment_id: str) -> None:
        self.model_instance_id = model_instance_id
        return super().delete(model_deployment_id)

    @arg("filename", help="Model request")
    def execute(self, model_instance_id: str, model_deployment_id: str, filename: str) -> EntityT:
        self.model_instance_id = model_instance_id
        request_dict = read_json_file(filename)
        model_instance_commands = ModelInstanceCommands(self._client_factory)
        model_instance = model_instance_commands.get(model_instance_id)
        model_response_cls = MODEL_SCHEMAS[model_instance.model_type][1]
        parsed_request = model_instance_commands._validate_request(model_instance, request_dict)
        sub_path = f"{self._get_api_subpath_prefix()}/{model_deployment_id}/execute"
        collection = self._get_collection_instance()
        response = collection._post(
            sub_path=sub_path,
            request=parsed_request,
            timeout=10 * 60,  # 10 minutes
        )
        assert response.status_code == 200
        return model_response_cls(**response.json())


class ModelGroupCommands(CollectionCRUDCommandsWithUpdate[ModelGroup, ModelGroupRequest]):
    command_group_name = "model-group"

    def __init__(
        self,
        client_factory: EGPClientFactory,
    ):
        super().__init__(client_factory, ModelGroup, ModelGroupRequest, ModelGroupCollection)


class ModelTemplateCommands(
    CollectionCRUDCommandsForImmutable[ModelTemplate, ModelTemplateRequest]
):
    command_group_name = "model-template"

    def __init__(
        self,
        client_factory: EGPClientFactory,
    ):
        super().__init__(
            client_factory, ModelTemplate, ModelTemplateRequest, ModelTemplateCollection
        )

    def _transform_entity_json(self, entity_dict: Dict[str, Any]) -> Dict[str, Any]:
        if entity_dict.get("vendor_configuration") is not None:
            entity_dict["model_vendor"] = "LAUNCH"
        return entity_dict

    def list(self) -> Generator[EntityT, None, None]:
        set_formatting_options(
            FormattingOptions(table_columns=["id", "name", "model_type", "created_at"])
        )
        return super().list()

    def show_model_schemas(self):
        return [
            {
                "model_type": model_type.value,
                "request_schema": schemas[0].schema(),
                "response_schema": schemas[1].schema(),
            }
            for (model_type, schemas) in MODEL_SCHEMAS.items()
        ]


class UserCommands:
    command_group_name = "user"
    command_group_title = "User management and account information"

    def __init__(
        self,
        client_factory: EGPClientFactory,
    ):
        self.client_factory = client_factory

    def whoami(self) -> UserInfoResponse:
        return self.client_factory.get_client().user_info()


class FineTuningJobCommands(
    CollectionCRUDCommandsForImmutable[FineTuningJob, FineTuningJobRequest]
):
    command_group_name = "fine-tuning-job"
    command_group_title = "Manage model fine tuning jobs"

    def __init__(
        self,
        client_factory: EGPClientFactory,
    ):
        super().__init__(
            client_factory, FineTuningJob, FineTuningJobRequest, FineTuningJobCollection
        )

    @arg("filename", nargs="?", help="file containing fine tuning job configuration")
    @arg("--base-model-id", help="Model instance id of base model to fine tune")
    @arg("--training-dataset-id", help="Training dataset")
    @arg("--validation-dataset-id", help="Validation dataset")
    def create(
        self,
        filename: str,
        base_model_id: Optional[str] = None,
        training_dataset_id: Optional[str] = None,
        validation_dataset_id: Optional[str] = None,
    ) -> EntityT:
        request_dict = read_json_file(filename) if filename else {}
        if base_model_id:
            request_dict["base_model_id"] = base_model_id
        if training_dataset_id:
            request_dict["training_dataset_id"] = training_dataset_id
        if validation_dataset_id:
            request_dict["validation_dataset_id"] = validation_dataset_id
        if request_dict.get("base_model_id") is None:
            raise CommandError("base_model_id must be set")
        if request_dict.get("training_dataset_id") is None:
            raise CommandError("training_dataset_id must be set")
        return self._create(request_dict)

    def events(self, fine_tuning_job_id: str) -> EntityT:
        sub_path = f"{self._get_api_subpath_prefix()}/{fine_tuning_job_id}/events"
        collection = self._get_collection_instance()
        response = collection._get(
            sub_path=sub_path,
            timeout=10 * 60,  # 10 minutes
        )
        assert response.status_code == 200
        for entity_dict in response.json():
            yield FineTuningJobEvent(**entity_dict)


class TrainingDatasetCommands(
    CollectionCRUDCommandsForImmutable[TrainingDataset, TrainingDatasetRequest]
):
    """
    Upload training and evaluation datasets to SGP for use in fine tuning model jobs.
    """

    command_group_name = "training-dataset"
    command_group_title = "Datasets for model fine tuning jobs"

    def __init__(
        self,
        client_factory: EGPClientFactory,
    ):
        super().__init__(
            client_factory, TrainingDataset, TrainingDatasetRequest, TrainingDatasetCollection
        )

    @arg("filename", help="training dataset file to upload")
    @arg(
        "--schema-type",
        type=str,
        choices=list([v.value for v in TrainingDatasetORMSchemaTypeEnum._member_map_.values()]),
        default=TrainingDatasetORMSchemaTypeEnum.GENERATION.value,
        help="Dataset type",
    )
    def create(
        self,
        filename: str,
        schema_type: Optional[str] = TrainingDatasetORMSchemaTypeEnum.GENERATION.value,
    ) -> EntityT:
        return TrainingDatasetCollection(self._client_factory.get_client()).create(
            file_name=filename, schema_type=TrainingDatasetORMSchemaTypeEnum(schema_type)
        )

    def list(self) -> Generator[EntityT, None, None]:
        set_formatting_options(
            FormattingOptions(
                table_columns=["id", "data_source", "created_at", "created_by_user_id"]
            )
        )
        return super().list()



from typing import Any, List, Optional, Dict

import httpx
from scale_egp.sdk.enums import ModelType, ModelVendor

from scale_egp.sdk.types.models import (
    ModelInstance,
    ModelDeployment,
    ModelDeploymentRequest,
    ModelInstanceRequest,
    BaseModelRequest,
    BaseModelResponse,
)
from scale_egp.sdk.constants.model_schemas import MODEL_SCHEMAS
from scale_egp.utils.api_utils import APIEngine
from scale_egp.utils.model_utils import dict_without_none_values, make_partial_model
from scale_egp.sdk.models.model_vendor_configuration import DeploymentVendorConfiguration


PartialModelAliasRequest = make_partial_model(ModelInstanceRequest)


class ModelInstanceCollection(APIEngine):
    """
    Collections class for SGP Models.
    """

    _sub_path = "v3/models"

    # START sub collections
    def deployments(self) -> "ModelDeploymentCollection":
        """
        Returns a ModelDeploymentCollection for deployments associated with this model.
        """
        return ModelDeploymentCollection(self._api_client)

    # END sub collections

    def create(
        self,
        name: str,
        model_type: ModelType,
        model_group_id: Optional[str] = None,
        model_vendor: Optional[ModelVendor] = None,
        model_template_id: Optional[str] = None,
        base_model_id: Optional[str] = None,
        base_model_metadata: Optional[Dict[str, Any]] = None,
        account_id: Optional[str] = None,
        model_card: Optional[str] = None,
        training_data_card: Optional[str] = None,
    ) -> ModelInstance:
        """
        Create a new SGP Model.

        Returns:
            The created Model.
        """
        response = self._post(
            sub_path=self._sub_path,
            request=ModelInstanceRequest(
                name=name,
                model_type=model_type,
                model_vendor=model_vendor,
                model_template_id=model_template_id,
                base_model_id=base_model_id,
                base_model_metadata=base_model_metadata,
                account_id=account_id or self._api_client.account_id,
                model_group_id=model_group_id,
                model_card=model_card,
                training_data_card=training_data_card,
            ),
        )
        return ModelInstance.from_dict(response.json())

    def get(
        self,
        id: str,
    ) -> ModelInstance:
        """
        Get a Model by ID.

        Returns:
            The Model.
        """
        response = self._get(
            sub_path=f"{self._sub_path}/{id}",
        )
        return ModelInstance.from_dict(response.json())

    def update(
        self,
        id: str,
        *,
        name: Optional[str] = None,
        model_template_id: Optional[str] = None,
        base_model_id: Optional[str] = None,
        model_creation_parameters: Optional[Dict[str, Any]] = None,
    ) -> ModelInstance:
        """
        Update a Model by ID.

        Returns:
            The updated Model.
        """
        response = self._patch(
            sub_path=f"{self._sub_path}/{id}",
            request=PartialModelAliasRequest(
                **dict_without_none_values(
                    dict(
                        name=name,
                        model_template_id=model_template_id,
                        base_model_id=base_model_id,
                        model_creation_parameters=model_creation_parameters,
                    ),
                )
            ),
        )
        return ModelInstance.from_dict(response.json())

    def delete(
        self,
        id: str,
    ) -> bool:
        """
        Delete a model by ID.

        Returns:
            True if the model was successfully deleted.
        """
        response = self._delete(
            sub_path=f"{self._sub_path}/{id}",
        )
        return response.status_code == httpx.codes.ok

    def list(
        self,
    ) -> List[ModelInstance]:
        """
        List all models that the user has access to.

        Returns:
            A list of Studio Projects.
        """
        response = self._get(
            sub_path=self._sub_path,
        )
        return [ModelInstance.from_dict(model) for model in response.json()]


class ModelDeploymentCollection(APIEngine):
    _sub_path = "v3/models/{model_id}/deployments"

    def create(
        self,
        model: ModelInstance,
        name: str,
        model_creation_parameters: Optional[Dict[str, Any]] = None,
        vendor_configuration: Optional[DeploymentVendorConfiguration] = None,
        account_id: Optional[str] = None,
    ) -> ModelDeployment:
        """
        Create a new ModelDeployment.

        Args:
            model: The Model to associate the ModelDeployment with.

        Returns:
            The newly created ModelDeployment.
        """
        response = self._post(
            sub_path=self._sub_path.format(model_id=model.id),
            request=ModelDeploymentRequest(
                name=name,
                model_creation_parameters=model_creation_parameters,
                vendor_configuration=vendor_configuration,
                model_id=model.id,
                account_id=account_id or self._api_client.account_id,
            ),
        )
        return ModelDeployment.from_dict(response.json())

    def get(
        self,
        id: str,
        model: ModelInstance,
    ) -> ModelDeployment:
        """
        Get a ModelDeployment by ID.

        Args:
            id: The ID of the ModelDeployment.
            model: The Model to associate the ModelDeployment with.

        Returns:
            The ModelDeployment.
        """
        response = self._get(
            sub_path=f"{self._sub_path.format(model_id=model.id)}/{id}",
        )
        return ModelDeployment.from_dict(response.json())

    # TODO <cathy-scale>: update?

    def delete(
        self,
        id: str,
        model: ModelInstance,
    ) -> bool:
        """
        Delete a ModelDeployment.

        Args:
            id: The ID of the ModelDeployment.
            model: The Model to associate the ModelDeployment with.
        """
        response = self._delete(
            sub_path=f"{self._sub_path.format(model_id=model.id)}/{id}",
        )
        return response.status_code == httpx.codes.ok

    def list(
        self,
        model: ModelInstance,
    ) -> List[ModelDeployment]:
        """
        List ModelDeployment.

        Args:
            model: The Model to associate the ModelDeployment with.

        Returns:
            A list of ModelDeployment.
        """
        response = self._get(
            sub_path=self._sub_path.format(model_id=model.id),
        )
        return [ModelDeployment.from_dict(deployment) for deployment in response.json()]

    def execute(
        self, id: str, model: ModelInstance, request: BaseModelRequest
    ) -> BaseModelResponse:
        """
        Execute the specified model deployment with the given request.

        Returns:
            The model deployment's response.
        """
        # TODO: verify model_request_parameters matches model template's
        #  model_request_parameters_schema if set
        model_request_cls, model_response_cls = MODEL_SCHEMAS[model.model_type]
        assert isinstance(request, model_request_cls)
        response = self._post(
            sub_path=f"{self._sub_path.format(model_id=model.id)}/{id}/execute",
            request=request,
            timeout=10 * 60,  # 10 minutes
        )
        return model_response_cls(**response.json())



from enum import Enum
from typing import Dict, List, Literal, Optional

from pydantic import Field

try:
    from egp_api_backend.server.api.utils.model_utils import BaseModel
except Exception:
    from pydantic import BaseModel

from .fine_tuning_jobs import TrainingDatasetORMSchemaTypeEnum
from .model_enums import ModelEndpointType, ModelVendor


class GPUType(str, Enum):
    # Supported GPU models according to
    # https://github.com/scaleapi/launch-python-client/blob/794089b9ed58330a7ecac02eb5060bcc4ae3d409/launch/client.py#L1470-L1471
    NVIDIA_TESLA_T4 = "nvidia-tesla-t4"
    NVIDIA_AMPERE_A10 = "nvidia-ampere-a10"
    NVIDIA_AMPERE_A100 = "nvidia-ampere-a100"
    NVIDIA_AMPERE_A100e = "nvidia-ampere-a100e"


class CommonBundleConfiguration(BaseModel):
    registry: str
    image: str
    tag: str
    # Note that the command field is mandatory since we're using create_model_bundle_from_streaming_enhanced_runnable_image_v2()
    # Omitting "command" triggers this error:
    # https://github.com/scaleapi/llm-engine/blob/53a1918ef3568b674b59a4e4e772501a7e1a1d69/model-engine/model_engine_server/domain/use_cases/model_endpoint_use_cases.py#L240
    command: List[str] = Field(default_factory=list)
    env: Dict[str, str] = Field(default_factory=dict)


# The fields in CreateModelBundleConfiguration(BaseModel) are based on the arguments to
# create_model_bundle_from_streaming_enhanced_runnable_image_v2() in the Launch python client:
# https://github.com/scaleapi/launch-python-client/blob/794089b9ed58330a7ecac02eb5060bcc4ae3d409/launch/client.py#L752
# Differences between this class and the function arguments:
# * Separate fields for docker repository and image name.
# * Omitted fields: tag, healthcheck_route, predict_route, metadata - for these fields EGP user will need to
#   use the defaults provided by Launch.
#
class ModelBundleConfiguration(CommonBundleConfiguration):
    streaming_command: Optional[List[str]] = Field(None)
    readiness_initial_delay_seconds: int = Field(120)
    healthcheck_route: str = Field("/readyz")
    predict_route: str = Field("/predict")
    streaming_predict_route: Optional[str] = Field("/generate_streaming")

    @property
    def full_repository_name(self):
        return "/".join([self.registry, self.image])


# Autoscaling options which can be updated on a per-deployment basis, overriding the model configuration
# in the model template.
class LaunchAutoscalingConfiguration(BaseModel):
    # By default, we create model endpoints with min_workers = 0 so unused model endpoints can be autoscaled down to
    # 0 workers, costing nothing.
    min_workers: int = Field(0)
    max_workers: int = Field(1)
    per_worker: int = Field(
        10,
        # from: https://github.com/scaleapi/launch-python-client/blob/794089b9ed58330a7ecac02eb5060bcc4ae3d409/launch/client.py#L1444-L1465
        description="""
The maximum number of concurrent requests that an individual worker can
service. Launch automatically scales the number of workers for the endpoint so that
each worker is processing ``per_worker`` requests, subject to the limits defined by
``min_workers`` and ``max_workers``.

- If the average number of concurrent requests per worker is lower than
``per_worker``, then the number of workers will be reduced. - Otherwise,
if the average number of concurrent requests per worker is higher than
``per_worker``, then the number of workers will be increased to meet the elevated
traffic.

Here is our recommendation for computing ``per_worker``:

1. Compute ``min_workers`` and ``max_workers`` per your minimum and maximum
throughput requirements. 2. Determine a value for the maximum number of
concurrent requests in the workload. Divide this number by ``max_workers``. Doing
this ensures that the number of workers will "climb" to ``max_workers``.
""".strip(),
    )


class RequiredResources(BaseModel):
    cpus: int = Field(3)
    memory: str = Field("8Gi")
    storage: str = Field("16Gi")
    gpus: int = Field(0)
    gpu_type: Optional[GPUType] = Field(None)


# The fields in CreateModelEndpointConfig are copied from the arguments of the Launch client's create_model_endpoint()
# https://github.com/scaleapi/launch-python-client/blob/794089b9ed58330a7ecac02eb5060bcc4ae3d409/launch/client.py#L1391
class CreateModelEndpointConfig(LaunchAutoscalingConfiguration, RequiredResources):
    endpoint_type: ModelEndpointType = Field(ModelEndpointType.ASYNC)
    high_priority: Optional[bool] = Field(False)


class FineTuningBundleConfiguration(CommonBundleConfiguration):
    # Information for the call to create a fine tuning batch job image
    # For example, see: https://github.com/scaleapi/models/blob/19b8c83e88962ac6b6a9d52b998c35345f862e80/enterprise/egp_finetuning/docker_utils/create_docker_bundle.py#L104-L142
    # Note that some
    mount_location: Optional[str] = Field(
        default="/workspace/launch_specific/config.json",
        description="The filesystem location where the fine tuning job's configuration will be available when it is started.",
    )
    training_dataset_schema_type: Optional[TrainingDatasetORMSchemaTypeEnum] = Field(
        default=None, description="Optionally set required training and validation dataset schema"
    )
    resources: Optional[RequiredResources] = Field(default_factory=RequiredResources)


class LaunchVendorConfiguration(BaseModel):
    """
    Configuration for launching a model using the Launch service which is an internal and
    self-hosted service developed by Scale that deploys models on Kubernetes.

    Attributes:
        vendor: The vendor of the model template
        bundle_config: The bundle configuration of the model template
        endpoint_config: The endpoint configuration of the model template
    """

    # this field is required for forward compatibility (other providers will have different "vendor" fields)
    vendor: Literal[ModelVendor.LAUNCH] = Field(ModelVendor.LAUNCH)
    bundle_config: ModelBundleConfiguration
    endpoint_config: Optional[CreateModelEndpointConfig] = Field(
        default_factory=CreateModelEndpointConfig
    )
    fine_tuning_job_bundle_config: Optional[FineTuningBundleConfiguration] = Field(None)


class DeploymentVendorConfiguration(LaunchAutoscalingConfiguration):
    # this field is required for forward compatibility (other providers will have different "vendor" fields)
    vendor: Literal[ModelVendor.LAUNCH] = Field(ModelVendor.LAUNCH)


# Model vendor configuration only necessary for Launch models currently
ModelVendorConfiguration = LaunchVendorConfiguration



from __future__ import annotations

from typing import Optional, Dict, Any, List

from pydantic import Field

from scale_egp.sdk.types.embeddings import EmbeddingConfig
from scale_egp.utils.model_utils import Entity, BaseModel


class KnowledgeBase(Entity):
    """
    A data model representing a knowledge base.

    Attributes:
        knowledge_base_id: The unique ID of the knowledge base
        knowledge_base_name: The name of the knowledge base
        embedding_config: The embedding configuration
        metadata: Metadata associated with the knowledge base
        created_at: The timestamp at which the knowledge base was created
        updated_at: The timestamp at which the knowledge base was last updated
    """

    knowledge_base_id: str
    knowledge_base_name: str
    embedding_config: EmbeddingConfig
    metadata: Optional[Dict[str, Any]]
    created_at: str
    updated_at: Optional[str]


class KnowledgeBaseRequest(BaseModel):
    knowledge_base_name: str = Field(..., description="A unique name for the knowledge base")
    embedding_config: EmbeddingConfig = Field(description="The configuration of the embedding")
    account_id: Optional[str] = Field(
        description="Account to create knowledge base in. If you have access to more than one "
                    "account, you must specify an account_id"
    )
    metadata: Optional[Dict[str, Any]] = Field(
        description="Metadata associated with the knowledge base"
    )


class KnowledgeBaseResponse(BaseModel):
    knowledge_base_id: str = Field(..., description="The unique ID of the created knowledge base")


class ListKnowledgeBasesResponse(BaseModel):
    items: List[KnowledgeBase] = Field(
        description="A list of the names and IDs, embedding configurations, metadata, created and "
                    "updated dates of your knowledge bases"
    )



from __future__ import annotations

from typing import Optional, Dict, Any, Union

from pydantic import Field, validator, ValidationError

from scale_egp.sdk.enums import TestCaseSchemaType, ExtraInfoSchema
from scale_egp.utils.model_utils import Entity, BaseModel


class ExtraInfo(BaseModel):
    """
    A data model representing the extra info of a generation model.

    Attributes:
        schema_type: The schema type of the extra info
        info: The content of the extra info. This must match the schema type.
    """

    schema_type: ExtraInfoSchema
    info: str

    @validator("info")
    def info_matches_schema_type(cls, info, values):
        schema_type = values.get("schema_type")
        if schema_type == ExtraInfoSchema.STRING:
            if not isinstance(info, str):
                raise ValueError("schema_type STRING requires info to be a string")
        else:
            raise ValueError(f"Unknown schema_type: {schema_type}")
        return info


class GenerationTestCaseData(BaseModel):
    """
    A data model representing the data of a Testcase with a GENERATION schema type.

    Attributes:
        input: The input to the generation model
        expected_output: The expected output of the generation model
        expected_extra_info: The expected extra info of the generation model
    """

    input: str
    expected_output: Optional[str] = ""
    expected_extra_info: Optional[ExtraInfo] = ExtraInfo(
        schema_type=ExtraInfoSchema.STRING,
        info="",
    )


class TestCase(Entity):
    """
    A data model representing a test case.

    Attributes:
        id: The ID of the test case
        evaluation_dataset_id: The ID of the evaluation dataset that the test case belongs to
        schema_type: The schema type of the test case
        test_case_data: The data of the test case. In general, this can be thought of as the
            data to evaluate an application against.
    """

    id: str
    evaluation_dataset_id: str
    schema_type: TestCaseSchemaType
    test_case_data: Union[GenerationTestCaseData]


class TestCaseRequest(BaseModel):
    schema_type: TestCaseSchemaType
    test_case_data: Union[GenerationTestCaseData]
    account_id: Optional[str] = Field(
        description="Account to create knowledge base in. If you have access to more than one "
                    "account, you must specify an account_id"
    )

    @validator("test_case_data")
    @classmethod
    def test_case_data_matches_schema_type(cls, test_case_data, values):
        schema_type = values.get("schema_type")
        TestCaseSchemaValidator.validate(schema_type, test_case_data)
        return test_case_data


class TestCaseSchemaValidator:
    TEST_CASE_SCHEMA_TO_DATA_TYPE = {
        TestCaseSchemaType.GENERATION: GenerationTestCaseData,
    }

    @classmethod
    def validate(cls, schema_type: TestCaseSchemaType, data: Dict[str, Any]):
        try:
            model = cls.TEST_CASE_SCHEMA_TO_DATA_TYPE[schema_type]
            model.from_dict(data)
        except ValidationError as e:
            raise ValueError(f"Test case data does not match schema type {schema_type}") from e

    @classmethod
    def get_model(cls, schema_type: TestCaseSchemaType):
        return cls.TEST_CASE_SCHEMA_TO_DATA_TYPE[schema_type]

    @classmethod
    def dict_to_model(cls, schema_type: TestCaseSchemaType, data: Dict[str, Any]):
        return cls.TEST_CASE_SCHEMA_TO_DATA_TYPE[schema_type].from_dict(data)



from enum import Enum


class ModelState(str, Enum):
    """
    An enum representing the different types of model states supported.

    Attributes:
        ENABLED: Denotes that the model is enabled.
        PENDING: Denotes that the model is pending.
        DISABLED: Denotes that the model is disabled.
    """

    ENABLED = "ENABLED"
    PENDING = "PENDING"
    DISABLED = "DISABLED"


class ModelVendor(str, Enum):
    """
    An enum representing the different types of model vendors supported.

    Attributes:
        OPENAI: Denotes that the model vendor is OpenAI.
        COHERE: Denotes that the model vendor is Cohere.
        ANTHROPIC: Denotes that the model vendor is Anthropic.
        LLMENGINE: Denotes that the model vendor is LLM Engine.
        OTHER: Denotes that the model vendor is Other.
    """

    OPENAI = "OPENAI"
    COHERE = "COHERE"
    ANTHROPIC = "ANTHROPIC"
    LAUNCH = "LAUNCH"
    OTHER = "OTHER"


class ModelEndpointType(str, Enum):
    """
    An enum representing the different types of model endpoint types supported.

    Attributes:
        SYNC: Denotes that the model endpoint type is sync.
        ASYNC: Denotes that the model endpoint type is async.
        STREAMING: Denotes that the model endpoint type is streaming.
        BATCH: Denotes that the model endpoint type is batch.
    """

    SYNC = "SYNC"
    ASYNC = "ASYNC"
    STREAMING = "STREAMING"
    BATCH = "BATCH"


class ModelType(str, Enum):
    """
    An enum representing the different types of models supported.

    Attributes:
        COMPLETION: Denotes that the model type is completion.
        CHAT_COMPLETION: Denotes that the model type is chat completion.
        AGENT: Denotes that the model type is agent.
        EMBEDDING: Denotes that the model type is embedding.
        RERANKING: Denotes that the model type is reranking.
    """

    COMPLETION = "COMPLETION"
    CHAT_COMPLETION = "CHAT_COMPLETION"
    AGENT = "AGENT"
    EMBEDDING = "EMBEDDING"
    RERANKING = "RERANKING"



from __future__ import annotations

from typing import Optional

from pydantic import Field

from scale_egp.utils.model_utils import BaseModel, Entity


class ApplicationSpecRequest(BaseModel):
    name: str
    description: str
    account_id: Optional[str] = Field(
        description="Account to create application spec in. If you have access to more than one "
                    "account, you must specify an account_id"
    )


class ApplicationSpec(Entity):
    """
    A data model representing an Application Spec.

    Attributes:
        id: The ID of the application spec
        name: The name of the application
        description: The description of the application

    """

    id: str
    name: str
    description: str



from __future__ import annotations

from datetime import datetime
from typing import Optional

from pydantic import Field

from scale_egp.utils.model_utils import Entity, BaseModel


class StudioProject(Entity):
    """
    A data model representing a Studio project.

    Attributes:
        id: The ID of the Studio project
        name: The name of the Studio project
        description: The description of the Studio project
        created_at: The time the Studio project was created
    """

    id: str
    name: str
    description: str
    created_at: datetime


class StudioProjectRequest(BaseModel):
    name: str
    description: str
    studio_api_key: str
    account_id: Optional[str] = Field(
        description="Account to create knowledge base in. If you have access to more than one "
                    "account, you must specify an account_id"
    )
